<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 08/25/2003 *
 * ===================                          Last change: 11/29/2005 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : functions.php                                    *
 * -------------------------------------------------------------------- *
 * Short description : Many non-database functions (also file access)   *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Viele Nicht-Datenbank-Funktionen                 *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2011 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// Init fatal message array
function initFatalMessages () {
	$GLOBALS['fatal_messages'] = array();
}

// Getter for whole fatal error messages
function getFatalArray () {
	return $GLOBALS['fatal_messages'];
}

// Add a fatal error message to the queue array
function addFatalMessage ($F, $L, $message, $extra = '') {
	if (is_array($extra)) {
		// Multiple extras for a message with masks
		$message = call_user_func_array('sprintf', $extra);
	} elseif (!empty($extra)) {
		// $message is text with a mask plus extras to insert into the text
		$message = sprintf($message, $extra);
	}

	// Add message to $GLOBALS['fatal_messages']
	$GLOBALS['fatal_messages'][] = $message;

	// Log fatal messages away
	logDebugMessage($F, $L, 'Fatal error message: ' . $message);
}

// Getter for total fatal message count
function getTotalFatalErrors () {
	// Init count
	$count = '0';

	// Do we have at least the first entry?
	if (!empty($GLOBALS['fatal_messages'][0])) {
		// Get total count
		$count = count($GLOBALS['fatal_messages']);
	} // END - if

	// Return value
	return $count;
}

// Send mail out to an email address
function sendEmail ($toEmail, $subject, $message, $isHtml = 'N', $mailHeader = '') {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'toEmail=' . $toEmail . ',subject=' . $subject . ',isHtml=' . $isHtml);
	// Empty parameters should be avoided, so we need to find them
	if (empty($isHtml)) {
		// isHtml is empty
		debug_report_bug(__FUNCTION__, __LINE__, 'isHtml is empty.');
	} // END - if

	// Set from header
	if ((!isInString('@', $toEmail)) && ($toEmail > 0)) {
		// Does the user exist?
		if ((isExtensionActive('user')) && (fetchUserData($toEmail))) {
			// Get the email
			$toEmail = getUserData('email');
		} else {
			// Set webmaster
			$toEmail = getWebmaster();
		}
	} elseif ($toEmail == '0') {
		// Is the webmaster!
		$toEmail = getWebmaster();
	}
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'TO=' . $toEmail);

	// Check for PHPMailer or debug-mode
	if ((!checkPhpMailerUsage()) || (isDebugModeEnabled())) {
		// Prefix is '' for text mails
		$prefix = '';

		// Is HTML?
		if ($isHtml == 'Y') {
			// Set prefix
			$prefix = 'html_';
		} // END - if

		// Not in PHPMailer-Mode
		if (empty($mailHeader)) {
			// Load email header template
			$mailHeader = loadEmailTemplate($prefix . 'header');
		} else {
			// Append header
			$mailHeader .= loadEmailTemplate($prefix . 'header');
		}
	} // END - if

	// Debug mode enabled?
	if (isDebugModeEnabled()) {
		// In debug mode we want to display the mail instead of sending it away so we can debug this part
		outputHtml('<pre>
Headers : ' . htmlentities(utf8_decode(trim($mailHeader))) . '
To      : ' . htmlentities(utf8_decode($toEmail)) . '
Subject : ' . htmlentities(utf8_decode($subject)) . '
Message : ' . htmlentities(utf8_decode($message)) . '
</pre>');

		// This is always fine
		return true;
	} elseif (!empty($toEmail)) {
		// Send Mail away
		return sendRawEmail($toEmail, $subject, $message, $mailHeader);
	} elseif ($isHtml != 'Y') {
		// Problem detected while sending a mail, forward it to admin
		return sendRawEmail(getWebmaster(), '[PROBLEM:]' . $subject, $message, $mailHeader);
	}

	// Why did we end up here? This should not happen
	debug_report_bug(__FUNCTION__, __LINE__, 'Ending up: template=' . $template);
}

// Check to use wether legacy mail() command or PHPMailer class
// @TODO Rewrite this to an extension 'smtp'
// @private
function checkPhpMailerUsage() {
	return ((getConfig('SMTP_HOSTNAME') != '') && (getConfig('SMTP_USER') != ''));
}

// Send out a raw email with PHPMailer class or legacy mail() command
function sendRawEmail ($toEmail, $subject, $message, $headers) {
	// Just compile all to put out all configs, etc.
	$eval  = '$toEmail = decodeEntities("' . escapeQuotes(doFinalCompilation(compileRawCode($toEmail), false)) . '"); ';
	$eval .= '$subject = decodeEntities("' . escapeQuotes(doFinalCompilation(compileRawCode($subject), false)) . '"); ';
	$eval .= '$headers = decodeEntities("' . escapeQuotes(doFinalCompilation(compileRawCode($headers), false)) . '"); ';

	// Do not decode entities in the message because we also send HTML mails through this function
	$eval .= '$message = "' . escapeQuotes(doFinalCompilation(compileRawCode($message), false)) . '";';

	// Run the final eval() command
	eval($eval);

	// Shall we use PHPMailer class or legacy mode?
	if (checkPhpMailerUsage()) {
		// Use PHPMailer class with SMTP enabled
		loadIncludeOnce('inc/phpmailer/class.phpmailer.php');
		loadIncludeOnce('inc/phpmailer/class.smtp.php');

		// get new instance
		$mail = new PHPMailer();

		// Set charset to UTF-8
		$mail->CharSet = 'UTF-8';

		// Path for PHPMailer
		$mail->PluginDir  = sprintf("%sinc/phpmailer/", getPath());

		$mail->IsSMTP();
		$mail->SMTPAuth   = true;
		$mail->Host       = getConfig('SMTP_HOSTNAME');
		$mail->Port       = 25;
		$mail->Username   = getConfig('SMTP_USER');
		$mail->Password   = getConfig('SMTP_PASSWORD');
		if (empty($headers)) {
			$mail->From = getWebmaster();
		} else {
			$mail->From = $headers;
		}
		$mail->FromName   = getMainTitle();
		$mail->Subject    = $subject;
		if ((isExtensionActive('html_mail')) && (secureString($message) != $message)) {
			$mail->Body       = $message;
			$mail->AltBody    = decodeEntities($message);
			$mail->WordWrap   = 70;
			$mail->IsHTML(true);
		} else {
			$mail->Body       = decodeEntities(strip_tags($message));
		}

		$mail->AddAddress($toEmail, '');
		$mail->AddReplyTo(getWebmaster(), getMainTitle());
		$mail->AddCustomHeader('Errors-To: ' . getWebmaster());
		$mail->AddCustomHeader('X-Loop: ' . getWebmaster());
		$mail->AddCustomHeader('Bounces-To: ' . getWebmaster());
		$mail->Send();

		// Has an error occured?
		if (!empty($mail->ErrorInfo)) {
			// Log message
			logDebugMessage(__FUNCTION__, __LINE__, 'Error while sending mail: ' . $mail->ErrorInfo);

			// Raise an error
			return false;
		} else {
			// All fine!
			return true;
		}
	} else {
		// Use legacy mail() command
		return mail($toEmail, $subject, decodeEntities($message), $headers);
	}
}

// Generate a password in a specified length or use default password length
function generatePassword ($length = '0', $exclude =  array()) {
	// Auto-fix invalid length of zero
	if ($length == '0') {
		$length = getPassLen();
	} // END - if

	// Initialize array with all allowed chars
	$ABC = explode(',', 'a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,0,1,2,3,4,5,6,7,8,9,-,+,_,/,.');

	// Exclude some entries
	$ABC = array_diff($ABC, $exclude);

	// Start creating password
	$PASS = '';
	for ($i = '0'; $i < $length; $i++) {
		$PASS .= $ABC[mt_rand(0, count($ABC) -1)];
	} // END - for

	// When the size is below 40 we can also add additional security by scrambling
	// it. Otherwise we may corrupt hashes
	if (strlen($PASS) <= 40) {
		// Also scramble the password
		$PASS = scrambleString($PASS);
	} // END - if

	// Return the password
	return $PASS;
}

// Generates a human-readable timestamp from the Uni* stamp
function generateDateTime ($time, $mode = '0') {
	// If the stamp is zero it mostly didn't "happen"
	if (($time == '0') || (is_null($time))) {
		// Never happend
		return '{--NEVER_HAPPENED--}';
	} // END - if

	// Filter out numbers
	$time = bigintval($time);

	// Is it cached?
	if (isset($GLOBALS[__FUNCTION__][$time][$mode])) {
		// Then use it
		return $GLOBALS[__FUNCTION__][$time][$mode];
	} // END - if

	// Detect language
	switch (getLanguage()) {
		case 'de': // German date / time format
			switch ($mode) {
				case '0': $ret = date("d.m.Y \u\m H:i \U\h\\r", $time); break;
				case '1': $ret = strtolower(date('d.m.Y - H:i', $time)); break;
				case '2': $ret = date('d.m.Y|H:i', $time); break;
				case '3': $ret = date('d.m.Y', $time); break;
				case '4': $ret = date('d.m.Y|H:i:s', $time); break;
				case '5': $ret = date('d-m-Y (l-F-T)', $time); break;
				case '6': $ret = date('Ymd', $time); break;
				case '7': $ret = date('Y-m-d H:i:s', $time); break; // Compatible with MySQL TIMESTAMP
				default:
					logDebugMessage(__FUNCTION__, __LINE__, sprintf("Invalid date mode %s detected.", $mode));
					break;
			} // END - switch
			break;

		default: // Default is the US date / time format!
			switch ($mode) {
				case '0': $ret = date('r', $time); break;
				case '1': $ret = strtolower(date('Y-m-d - g:i A', $time)); break;
				case '2': $ret = date('y-m-d|H:i', $time); break;
				case '3': $ret = date('y-m-d', $time); break;
				case '4': $ret = date('d.m.Y|H:i:s', $time); break;
				case '5': $ret = date('d-m-Y (l-F-T)', $time); break;
				case '6': $ret = date('Ymd', $time); break;
				case '7': $ret = date('Y-m-d H:i:s', $time); break; // Compatible with MySQL TIMESTAMP
				default:
					logDebugMessage(__FUNCTION__, __LINE__, sprintf("Invalid date mode %s detected.", $mode));
					break;
			} // END - switch
	} // END - switch

	// Store it in cache
	$GLOBALS[__FUNCTION__][$time][$mode] = $ret;

	// Return result
	return $ret;
}

// Translates Y/N to yes/no
function translateYesNo ($yn) {
	// Is it cached?
	if (!isset($GLOBALS[__FUNCTION__][$yn])) {
		// Default
		$GLOBALS[__FUNCTION__][$yn] = '??? (' . $yn . ')';
		switch ($yn) {
			case 'Y': $GLOBALS[__FUNCTION__][$yn] = '{--YES--}'; break;
			case 'N': $GLOBALS[__FUNCTION__][$yn] = '{--NO--}'; break;
			default:
				// Log unknown value
				logDebugMessage(__FUNCTION__, __LINE__, sprintf("Unknown value %s. Expected Y/N!", $yn));
				break;
		} // END - switch
	} // END - if

	// Return it
	return $GLOBALS[__FUNCTION__][$yn];
}

// Translates the american decimal dot into a german comma
function translateComma ($dotted, $cut = true, $max = '0') {
	// First, cast all to double, due to PHP changes
	$dotted = (double) $dotted;

	// Default is 3 you can change this in admin area "Settings -> Misc Options"
	if (!isConfigEntrySet('max_comma')) {
		setConfigEntry('max_comma', 3);
	} // END - if

	// Use from config is default
	$maxComma = getConfig('max_comma');

	// Use from parameter?
	if ($max > 0) {
		$maxComma = $max;
	} // END - if

	// Cut zeros off?
	if (($cut === true) && ($max == '0')) {
		// Test for commata if in cut-mode
		$com = explode('.', $dotted);
		if (count($com) < 2) {
			// Don't display commatas even if there are none... ;-)
			$maxComma = '0';
		} // END - if
	} // END - if

	// Debug log

	// Translate it now
	$translated = $dotted;
	switch (getLanguage()) {
		case 'de': // German language
			$translated = number_format($dotted, $maxComma, ',', '.');
			break;

		default: // All others
			$translated = number_format($dotted, $maxComma, '.', ',');
			break;
	} // END - switch

	// Return translated value
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'dotted=' . $dotted . ',translated=' . $translated . ',maxComma=' . $maxComma);
	return $translated;
}

// Translate Uni*-like gender to human-readable
function translateGender ($gender) {
	// Default
	$ret = '!' . $gender . '!';

	// Male/female or company?
	switch ($gender) {
		case 'M': // Male
		case 'F': // Female
		case 'C': // Company
			$ret = sprintf("{--GENDER_%s--}", $gender);
			break;

		default:
			// Please report bugs on unknown genders
			debug_report_bug(__FUNCTION__, __LINE__, sprintf("Unknown gender %s detected.", $gender));
			break;
	} // END - switch

	// Return translated gender
	return $ret;
}

// "Translates" the user status
function translateUserStatus ($status) {
	// Default status is unknown if something goes through
	$ret = '{--ACCOUNT_STATUS_UNKNOWN--}';

	// Generate message depending on status
	switch ($status) {
		case 'UNCONFIRMED':
		case 'CONFIRMED':
		case 'LOCKED':
			$ret = sprintf("{--ACCOUNT_STATUS_%s--}", $status);
			break;

		case '':
		case null:
			$ret = '{--ACCOUNT_STATUS_DELETED--}';
			break;

		default:
			// Please report all unknown status
			debug_report_bug(__FUNCTION__, __LINE__, sprintf("Unknown status %s(%s) detected.", $status, gettype($status)));
			break;
	} // END - switch

	// Return it
	return $ret;
}

// "Translates" 'visible' and 'locked' to a CSS class
function translateMenuVisibleLocked ($content, $prefix = '') {
	// Default is 'menu_unknown'
	$content['visible_css'] = $prefix . 'menu_unknown';

	// Translate 'visible' and keep an eye on the prefix
	switch ($content['visible']) {
		// Should be visible
		case 'Y': $content['visible_css'] = $prefix . 'menu_visible'  ; break;
		case 'N': $content['visible_css'] = $prefix . 'menu_invisible'; break;
		default:
			// Please report this
			debug_report_bug(__FUNCTION__, __LINE__, 'Unsupported visible value detected. content=<pre>' . print_r($content, true) . '</pre>');
			break;
	} // END - switch

	// Translate 'locked' and keep an eye on the prefix
	switch ($content['locked']) {
		// Should be locked
		case 'Y': $content['locked_css'] = $prefix . 'menu_locked'  ; break;
		case 'N': $content['locked_css'] = $prefix . 'menu_unlocked'; break;
		default:
			// Please report this
			debug_report_bug(__FUNCTION__, __LINE__, 'Unsupported locked value detected. content=<pre>' . print_r($content, true) . '</pre>');
			break;
	} // END - switch

	// Return the resulting array
	return $content;
}

// Generates an URL for the dereferer
function generateDerefererUrl ($url) {
	// Don't de-refer our own links!
	if (substr($url, 0, strlen(getUrl())) != getUrl()) {
		// De-refer this link
		$url = '{%url=modules.php?module=loader&amp;url=' . encodeString(compileUriCode($url)) . '%}';
	} // END - if

	// Return link
	return $url;
}

// Generates an URL for the frametester
function generateFrametesterUrl ($url) {
	// Prepare frametester URL
	$frametesterUrl = sprintf("{%%url=modules.php?module=frametester&amp;url=%s%%}",
		encodeString(compileUriCode($url))
	);

	// Return the new URL
	return $frametesterUrl;
}

// Count entries from e.g. a selection box
function countSelection ($array) {
	// Integrity check
	if (!is_array($array)) {
		// Not an array!
		debug_report_bug(__FUNCTION__, __LINE__, 'No array provided.');
	} // END - if

	// Init count
	$ret = '0';

	// Count all entries
	foreach ($array as $key => $selected) {
		// Is it checked?
		if (!empty($selected)) $ret++;
	} // END - foreach

	// Return counted selections
	return $ret;
}

// Generates a timestamp (some wrapper for mktime())
function makeTime ($hours, $minutes, $seconds, $stamp) {
	// Extract day, month and year from given timestamp
	$days   = getDay($stamp);
	$months = getMonth($stamp);
	$years  = getYear($stamp);

	// Create timestamp for wished time which depends on extracted date
	return mktime(
		$hours,
		$minutes,
		$seconds,
		$months,
		$days,
		$years
	);
}

// Redirects to an URL and if neccessarry extends it with own base URL
function redirectToUrl ($url, $allowSpider = true) {
	// Remove {%url=
	if (substr($url, 0, 6) == '{%url=') {
		$url = substr($url, 6, -2);
	} // END - if

	// Compile out codes
	eval('$url = "' . compileRawCode(encodeUrl($url)) . '";');

	// Default 'rel' value is external, nofollow is evil from Google and hurts the Internet
	$rel = ' rel="external"';

	// Do we have internal or external URL?
	if (substr($url, 0, strlen(getUrl())) == getUrl()) {
		// Own (=internal) URL
		$rel = '';
	} // END - if

	// Three different ways to debug...
	//* DEBUG: */ debug_report_bug(__FUNCTION__, __LINE__, 'URL=' . $url);
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'URL=' . $url);
	//* DEBUG: */ die($url);

	// We should not sent a redirect if headers are already sent
	if (!headers_sent()) {
		// Clear output buffer
		clearOutputBuffer();

		// Clear own output buffer
		$GLOBALS['output'] = '';

		// Load URL when headers are not sent
		sendRawRedirect(doFinalCompilation(str_replace('&amp;', '&', $url), false));
	} else {
		// Output error message
		loadInclude('inc/header.php');
		loadTemplate('redirect_url', false, str_replace('&amp;', '&', $url));
		loadInclude('inc/footer.php');
	}

	// Shut the mailer down here
	shutdown();
}

/************************************************************************
 *                                                                      *
 * Gaenderter Sortier-Algorythmus, $array wird nach dem Array (!)       *
 * $a_sort sortiert:                                                    *
 *                                                                      *
 * $array - Das 3-dimensionale Array, das paralell sortiert werden soll *
 * $a_sort - Array, das die Sortiereihenfolge der ersten Elementeben    *
 * $primary_key - Prim.rschl.ssel aus $a_sort, nach dem sortiert wird   *
 * $order - Sortiereihenfolge: -1 = a-Z, 0 = keine, 1 = Z-a             *
 * $nums - true = Als Zahlen sortieren, false = Als Zeichen sortieren   *
 *                                                                      *
 * $a_sort muss Elemente enthalten, deren Wert Schluessel von $array    *
 * sind... Klingt kompliziert, suchen Sie mal mein Beispiel, dann sehen *
 * Sie, dass es doch nicht so schwer ist! :-)                           *
 *                                                                      *
 ************************************************************************/
function array_pk_sort (&$array, $a_sort, $primary_key = '0', $order = -1, $nums = false) {
	$temporaryArray = $array;
	while ($primary_key < count($a_sort)) {
		foreach ($temporaryArray[$a_sort[$primary_key]] as $key => $value) {
			foreach ($temporaryArray[$a_sort[$primary_key]] as $key2 => $value2) {
				$match = false;
				if ($nums === false) {
					// Sort byte-by-byte (also numbers will be interpreted as chars! E.g.: "9" > "10")
					if (($key != $key2) && (strcmp(strtolower($temporaryArray[$a_sort[$primary_key]][$key]), strtolower($temporaryArray[$a_sort[$primary_key]][$key2])) == $order)) $match = true;
				} elseif ($key != $key2) {
					// Sort numbers (E.g.: 9 < 10)
					if (($temporaryArray[$a_sort[$primary_key]][$key] < $temporaryArray[$a_sort[$primary_key]][$key2]) && ($order == -1)) $match = true;
					if (($temporaryArray[$a_sort[$primary_key]][$key] > $temporaryArray[$a_sort[$primary_key]][$key2]) && ($order == 1))  $match = true;
				}

				if ($match) {
					// We have found two different values, so let's sort whole array
					foreach ($temporaryArray as $sort_key => $sort_val) {
						$t                       = $temporaryArray[$sort_key][$key];
						$temporaryArray[$sort_key][$key]  = $temporaryArray[$sort_key][$key2];
						$temporaryArray[$sort_key][$key2] = $t;
						unset($t);
					} // END - foreach
				} // END - if
			} // END - foreach
		} // END - foreach

		// Count one up
		$primary_key++;
	} // END - while

	// Write back sorted array
	$array = $temporaryArray;
}


//
// Deprecated : $length (still has one reference in this function)
// Optional   : $extraData
//
function generateRandomCode ($length, $code, $userid, $extraData = '') {
	// Build server string
	$server = $_SERVER['PHP_SELF'] . getEncryptSeparator() . detectUserAgent() . getEncryptSeparator() . getenv('SERVER_SOFTWARE') . getEncryptSeparator() . detectRealIpAddress() . getEncryptSeparator() . detectRemoteAddr();

	// Build key string
	$keys = getSiteKey() . getEncryptSeparator() . getDateKey();
	if (isConfigEntrySet('secret_key')) {
		$keys .= getEncryptSeparator() . getSecretKey();
	} // END - if
	if (isConfigEntrySet('file_hash')) {
		$keys .= getEncryptSeparator() . getFileHash();
	} // END - if
	$keys .= getEncryptSeparator() . getDateFromRepository();
	if (isConfigEntrySet('master_salt')) {
		$keys .= getEncryptSeparator() . getMasterSalt();
	} // END - if

	// Build string from misc data
	$data  = $code . getEncryptSeparator() . $userid . getEncryptSeparator() . $extraData;

	// Add more additional data
	if (isSessionVariableSet('u_hash')) {
		$data .= getEncryptSeparator() . getSession('u_hash');
	} // END - if

	// Add referral id, language, theme and userid
	$data .= getEncryptSeparator() . determineReferralId();
	$data .= getEncryptSeparator() . getLanguage();
	$data .= getEncryptSeparator() . getCurrentTheme();
	$data .= getEncryptSeparator() . getMemberId();

	// Calculate number for generating the code
	$a = $code + getConfig('_ADD') - 1;

	if (isConfigEntrySet('master_salt')) {
		// Generate hash with master salt from modula of number with the prime number and other data
		$saltedHash = generateHash(($a % getPrime()) . getEncryptSeparator() . $server . getEncryptSeparator() . $keys . getEncryptSeparator() . $data . getEncryptSeparator() . getDateKey() . getEncryptSeparator() . $a, getMasterSalt());
	} else {
		// Generate hash with "hash of site key" from modula of number with the prime number and other data
		$saltedHash = generateHash(($a % getPrime()) . getEncryptSeparator() . $server . getEncryptSeparator() . $keys . getEncryptSeparator() . $data . getEncryptSeparator() . getDateKey() . getEncryptSeparator() . $a, substr(sha1(getSiteKey()), 0, getSaltLength()));
	}

	// Create number from hash
	$rcode = hexdec(substr($saltedHash, getSaltLength(), 9)) / abs(getRandNo() - $a + sqrt(getConfig('_ADD'))) / pi();

	// At least 10 numbers shall be secure enought!
	if (isExtensionActive('other')) {
		$len = getCodeLength();
	} else {
		$len = $length;
	} // END - if

	if ($len == '0') {
		$len = 10;
	} // END - if

	// Cut off requested counts of number, but skip first digit (which is mostly a zero)
	$return = substr($rcode, (strpos($rcode, '.') + 1), $len);

	// Done building code
	return $return;
}

// Does only allow numbers
function bigintval ($num, $castValue = true, $abortOnMismatch = true) {
	//* DEBUG: */ debugOutput('[' . __FUNCTION__ . ':' . __LINE__ . '] ' . 'num=' . $num . ',castValue=' . intval($castValue) . ',abortOnMismatch=' . intval($abortOnMismatch) . ' - ENTERED!');
	// Filter all numbers out
	$ret = preg_replace('/[^0123456789]/', '', $num);

	// Shall we cast?
	if ($castValue === true) {
		// Cast to biggest numeric type
		$ret = (double) $ret;
	} // END - if

	// Has the whole value changed?
	if (('' . $ret . '' != '' . $num . '') && ($abortOnMismatch === true) && (!is_null($num))) {
		// Log the values
		debug_report_bug(__FUNCTION__, __LINE__, 'Problem with number found. ret[' . gettype($ret) . ']=' . $ret . ', num[' . gettype($num) . ']='. $num);
	} // END - if

	// Return result
	//* DEBUG: */ debugOutput('[' . __FUNCTION__ . ':' . __LINE__ . '] ' . 'num=' . $num . ',castValue=' . intval($castValue) . ',abortOnMismatch=' . intval($abortOnMismatch) . ',ret=' . $ret . ' - EXIT!');
	return $ret;
}

// Creates a Uni* timestamp from given selection data and prefix
function createEpocheTimeFromSelections ($prefix, $postData) {
	// Initial return value
	$ret = '0';

	// Do we have a leap year?
	$SWITCH = '0';
	$TEST = getYear() / 4;
	$M1   = getMonth();

	// If so and if current time is before 02/29 and estimated time is after 02/29 then add 86400 seconds (one day)
	if ((floor($TEST) == $TEST) && ($M1 == '02') && ($postData[$prefix . '_mo'] > '02'))  {
		$SWITCH = getOneDay();
	} // END - if

	// First add years...
	$ret += $postData[$prefix . '_ye'] * (31536000 + $SWITCH);

	// Next months...
	$ret += $postData[$prefix . '_mo'] * 2628000;

	// Next weeks
	$ret += $postData[$prefix . '_we'] * 604800;

	// Next days...
	$ret += $postData[$prefix . '_da'] * 86400;

	// Next hours...
	$ret += $postData[$prefix . '_ho'] * 3600;

	// Next minutes..
	$ret += $postData[$prefix . '_mi'] * 60;

	// And at last seconds...
	$ret += $postData[$prefix . '_se'];

	// Return calculated value
	return $ret;
}

// Creates a 'fancy' human-readable timestamp from a Uni* stamp
function createFancyTime ($stamp) {
	// Get data array with years/months/weeks/days/...
	$data = createTimeSelections($stamp, '', '', '', true);
	$ret = '';
	foreach ($data as $k => $v) {
		if ($v > 0) {
			// Value is greater than 0 "eval" data to return string
			$ret .= ', ' . $v . ' {%pipe,translateTimeUnit=' . $k . '%}';
			break;
		} // END - if
	} // END - foreach

	// Do we have something there?
	if (strlen($ret) > 0) {
		// Remove leading commata and space
		$ret = substr($ret, 2);
	} else {
		// Zero seconds
		$ret = '0 {--TIME_UNIT_SECOND--}';
	}

	// Return fancy time string
	return $ret;
}

// Taken from www.php.net isInStringIgnoreCase() user comments
function isEmailValid ($email) {
	// Check first part of email address
	$first = '[-a-z0-9!#$%&\'*+/=?^_<{|}~]+(\.[-a-zA-Z0-9!#$%&\'*+/=?^_<{|}~]+)*';

	//  Check domain
	$domain = '[a-z0-9-]+(\.[a-z0-9-]{2,5})+';

	// Generate pattern
	$regex = '@^' . $first . '\@' . $domain . '$@iU';

	// Return check result
	return preg_match($regex, $email);
}

// Function taken from user comments on www.php.net / function isInStringIgnoreCase()
function isUrlValid ($url, $compile = true) {
	// Trim URL a little
	$url = trim(urldecode($url));
	//* DEBUG: */ debugOutput($url);

	// Compile some chars out...
	if ($compile === true) {
		$url = compileUriCode($url, false, false, false);
	} // END - if
	//* DEBUG: */ debugOutput($url);

	// Check for the extension filter
	if (isExtensionActive('filter')) {
		// Use the extension's filter set
		return FILTER_VALIDATE_URL($url, false);
	} // END - if

	// If not installed, perform a simple test. Just make it sure there is always a http:// or
	// https:// in front of the URLs
	return isUrlValidSimple($url);
}

// Generate a hash for extra-security for all passwords
function generateHash ($plainText, $salt = '', $hash = true) {
	// Debug output
	//* DEBUG: */ debugOutput('plainText('.strlen($plainText).')=' . $plainText . ',salt('.strlen($salt).')=' . $salt . ',hash=' . intval($hash));

	// Is the required extension 'sql_patches' there and a salt is not given?
	// 123                            4                      43    3     4     432    2                  3             32    2                             3                32    2      3     3      21
	if (((isExtensionInstalledAndOlder('sql_patches', '0.3.6')) && (empty($salt))) || (!isExtensionActive('sql_patches')) || (!isExtensionInstalledAndNewer('other', '0.2.5')) || (strlen($salt) == 32)) {
		// Extension ext-sql_patches is missing/outdated so we hash the plain text with MD5
		if ($hash === true) {
			// Is plain password
			return md5($plainText);
		} else {
			// Is already a hash
			return $plainText;
		}
	} // END - if

	// Do we miss an arry element here?
	if (!isConfigEntrySet('file_hash')) {
		// Stop here
		debug_report_bug(__FUNCTION__, __LINE__, 'Missing file_hash in ' . __FUNCTION__ . '.');
	} // END - if

	// When the salt is empty build a new one, else use the first x configured characters as the salt
	if (empty($salt)) {
		// Build server string for more entropy
		$server = $_SERVER['PHP_SELF'] . getEncryptSeparator() . detectUserAgent() . getEncryptSeparator() . getenv('SERVER_SOFTWARE') . getEncryptSeparator() . detectRealIpAddress() . getEncryptSeparator() . detectRemoteAddr();

		// Build key string
		$keys   = getSiteKey() . getEncryptSeparator() . getDateKey() . getEncryptSeparator() . getSecretKey() . getEncryptSeparator() . getFileHash() . getEncryptSeparator() . getDateFromRepository() . getEncryptSeparator() . getMasterSalt();

		// Additional data
		$data = $plainText . getEncryptSeparator() . uniqid(mt_rand(), true) . getEncryptSeparator() . time();

		// Calculate number for generating the code
		$a = time() + getConfig('_ADD') - 1;

		// Generate SHA1 sum from modula of number and the prime number
		$sha1 = sha1(($a % getPrime()) . $server . getEncryptSeparator() . $keys . getEncryptSeparator() . $data . getEncryptSeparator() . getDateKey() . getEncryptSeparator() . $a);
		//* DEBUG: */ debugOutput('SHA1=' . $sha1.' ('.strlen($sha1).')<br />');
		$sha1 = scrambleString($sha1);
		//* DEBUG: */ debugOutput('Scrambled=' . $sha1.' ('.strlen($sha1).')<br />');
		//* DEBUG: */ $sha1b = descrambleString($sha1);
		//* DEBUG: */ debugOutput('Descrambled=' . $sha1b.' ('.strlen($sha1b).')<br />');

		// Generate the password salt string
		$salt = substr($sha1, 0, getSaltLength());
		//* DEBUG: */ debugOutput($salt.' ('.strlen($salt).')<br />');
	} else {
		// Use given salt
		//* DEBUG: */ debugOutput('salt=' . $salt);
		$salt = substr($salt, 0, getSaltLength());
		//* DEBUG: */ debugOutput('salt=' . $salt . '(' . strlen($salt) . '/' . getSaltLength() . ')<br />');

		// Sanity check on salt
		if (strlen($salt) != getSaltLength()) {
			// Not the same!
			debug_report_bug(__FUNCTION__, __LINE__, 'salt length mismatch! (' . strlen($salt) . '/' . getSaltLength() . ')');
		} // END - if
	}

	// Generate final hash (for debug output)
	$finalHash = $salt . sha1($salt . $plainText);

	// Debug output
	//* DEBUG: */ debugOutput('finalHash('.strlen($finalHash).')=' . $finalHash);

	// Return hash
	return $finalHash;
}

// Scramble a string
function scrambleString ($str) {
	// Init
	$scrambled = '';

	// Final check, in case of failure it will return unscrambled string
	if (strlen($str) > 40) {
		// The string is to long
		return $str;
	} elseif (strlen($str) == 40) {
		// From database
		$scrambleNums = explode(':', getPassScramble());
	} else {
		// Generate new numbers
		$scrambleNums = explode(':', genScrambleString(strlen($str)));
	}

	// Compare both lengths and abort if different
	if (strlen($str) != count($scrambleNums)) return $str;

	// Scramble string here
	//* DEBUG: */ debugOutput('***Original=' . $str.'***<br />');
	for ($idx = 0; $idx < strlen($str); $idx++) {
		// Get char on scrambled position
		$char = substr($str, $scrambleNums[$idx], 1);

		// Add it to final output string
		$scrambled .= $char;
	} // END - for

	// Return scrambled string
	//* DEBUG: */ debugOutput('***Scrambled=' . $scrambled.'***<br />');
	return $scrambled;
}

// De-scramble a string scrambled by scrambleString()
function descrambleString ($str) {
	// Scramble only 40 chars long strings
	if (strlen($str) != 40) return $str;

	// Load numbers from config
	$scrambleNums = explode(':', getPassScramble());

	// Validate numbers
	if (count($scrambleNums) != 40) return $str;

	// Begin descrambling
	$orig = str_repeat(' ', 40);
	//* DEBUG: */ debugOutput('+++Scrambled=' . $str.'+++<br />');
	for ($idx = 0; $idx < 40; $idx++) {
		$char = substr($str, $idx, 1);
		$orig = substr_replace($orig, $char, $scrambleNums[$idx], 1);
	} // END - for

	// Return scrambled string
	//* DEBUG: */ debugOutput('+++Original=' . $orig.'+++<br />');
	return $orig;
}

// Generated a "string" for scrambling
function genScrambleString ($len) {
	// Prepare array for the numbers
	$scrambleNumbers = array();

	// First we need to setup randomized numbers from 0 to 31
	for ($idx = 0; $idx < $len; $idx++) {
		// Generate number
		$rand = mt_rand(0, ($len - 1));

		// Check for it by creating more numbers
		while (array_key_exists($rand, $scrambleNumbers)) {
			$rand = mt_rand(0, ($len - 1));
		} // END - while

		// Add number
		$scrambleNumbers[$rand] = $rand;
	} // END - for

	// So let's create the string for storing it in database
	$scrambleString = implode(':', $scrambleNumbers);
	return $scrambleString;
}

// Generate an PGP-like encrypted hash of given hash for e.g. cookies
function encodeHashForCookie ($passHash) {
	// Return vanilla password hash
	$ret = $passHash;

	// Is a secret key and master salt already initialized?
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, intval(isExtensionInstalled('sql_patches')) . '/' . intval(isConfigEntrySet('_PRIME')) . '/' . intval(isConfigEntrySet('secret_key')) . '/' . intval(isConfigEntrySet('master_salt')));
	if ((isExtensionInstalled('sql_patches')) && (isConfigEntrySet('_PRIME')) && (isConfigEntrySet('secret_key')) && (isConfigEntrySet('master_salt'))) {
		// Only calculate when the secret key is generated
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, strlen($passHash) . '/' . strlen(getSecretKey()));
		if ((strlen($passHash) != 49) || (strlen(getSecretKey()) != 40)) {
			// Both keys must have same length so return unencrypted
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, strlen($passHash) . '!=49/' . strlen(getSecretKey()) . '!=40');
			return $ret;
		} // END - if

		$newHash = ''; $start = 9;
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'passHash=' . $passHash . '(' . strlen($passHash) . ')');
		for ($idx = 0; $idx < 20; $idx++) {
			$part1 = hexdec(substr($passHash, ($idx * 2) + (strlen($passHash) - strlen(getSecretKey())), 2));
			$part2 = hexdec(substr(getSecretKey(), $start, 2));
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'part1=' . $part1 . '/part2=' . $part2);
			$mod = dechex($idx);
			if ($part1 > $part2) {
				$mod = dechex(sqrt(($part1 - $part2) * getPrime() / pi()));
			} elseif ($part2 > $part1) {
				$mod = dechex(sqrt(($part2 - $part1) * getPrime() / pi()));
			}
			$mod = substr($mod, 0, 2);
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'part1=' . $part1 . '/part2=' . $part2 . '/mod=' . $mod . '(' . strlen($mod) . ')');
			$mod = str_repeat(0, (2 - strlen($mod))) . $mod;
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'mod(' . ($idx * 2) . ')=' . $mod . '*');
			$start += 2;
			$newHash .= $mod;
		} // END - for

		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $passHash . ',' . $newHash . ' (' . strlen($newHash) . ')');
		$ret = generateHash($newHash, getMasterSalt());
	} // END - if

	// Return result
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'ret=' . $ret . '');
	return $ret;
}

// Fix "deleted" cookies
function fixDeletedCookies ($cookies) {
	// Is this an array with entries?
	if ((is_array($cookies)) && (count($cookies) > 0)) {
		// Then check all cookies if they are marked as deleted!
		foreach ($cookies as $cookieName) {
			// Is the cookie set to "deleted"?
			if (getSession($cookieName) == 'deleted') {
				setSession($cookieName, '');
			} // END - if
		} // END - foreach
	} // END - if
}

// Checks if a given apache module is loaded
function isApacheModuleLoaded ($apacheModule) {
	// Check it and return result
	return (((function_exists('apache_get_modules')) && (in_array($apacheModule, apache_get_modules()))) || (!function_exists('apache_get_modules')));
}

// Get current theme name
function getCurrentTheme () {
	// The default theme is 'default'... ;-)
	$ret = 'default';

	// Do we have ext-theme installed and active?
	if (isExtensionActive('theme')) {
		// Call inner method
		$ret = getActualTheme();
	} // END - if

	// Return theme value
	return $ret;
}

// Generates an error code from given account status
function generateErrorCodeFromUserStatus ($status = '') {
	// If no status is provided, use the default, cached
	if ((empty($status)) && (isMember())) {
		// Get user status
		$status = getUserData('status');
	} // END - if

	// Default error code if unknown account status
	$errorCode = getCode('ACCOUNT_UNKNOWN');

	// Generate constant name
	$codeName = sprintf("ACCOUNT_%s", strtoupper($status));

	// Is the constant there?
	if (isCodeSet($codeName)) {
		// Then get it!
		$errorCode = getCode($codeName);
	} else {
		// Unknown status
		logDebugMessage(__FUNCTION__, __LINE__, sprintf("Unknown error status %s detected.", $status));
	}

	// Return error code
	return $errorCode;
}

// Back-ported from the new ship-simu engine. :-)
function debug_get_printable_backtrace () {
	// Init variable
	$backtrace = '<ol>';

	// Get and prepare backtrace for output
	$backtraceArray = debug_backtrace();
	foreach ($backtraceArray as $key => $trace) {
		if (!isset($trace['file'])) $trace['file'] = __FUNCTION__;
		if (!isset($trace['line'])) $trace['line'] = __LINE__;
		if (!isset($trace['args'])) $trace['args'] = array();
		$backtrace .= '<li class="debug_list"><span class="backtrace_file">' . basename($trace['file']) . '</span>:' . $trace['line'] . ', <span class="backtrace_function">' . $trace['function'] . '(' . count($trace['args']) . ')</span></li>';
	} // END - foreach

	// Close it
	$backtrace .= '</ol>';

	// Return the backtrace
	return $backtrace;
}

// A mail-able backtrace
function debug_get_mailable_backtrace () {
	// Init variable
	$backtrace = '';

	// Get and prepare backtrace for output
	$backtraceArray = debug_backtrace();
	foreach ($backtraceArray as $key => $trace) {
		if (!isset($trace['file'])) $trace['file'] = __FUNCTION__;
		if (!isset($trace['line'])) $trace['line'] = __LINE__;
		if (!isset($trace['args'])) $trace['args'] = array();
		$backtrace .= ($key+1) . '.:' . basename($trace['file']) . ':' . $trace['line'] . ', ' . $trace['function'] . '(' . count($trace['args']) . ")\n";
	} // END - foreach

	// Return the backtrace
	return $backtrace;
}

// Generates a ***weak*** seed
function generateSeed () {
	return microtime(true) * 100000;
}

// Converts a message code to a human-readable message
function getMessageFromErrorCode ($code) {
	$message = '';
	switch ($code) {
		case '': break;
		case getCode('LOGOUT_DONE')        : $message = '{--LOGOUT_DONE--}'; break;
		case getCode('LOGOUT_FAILED')      : $message = '<span class="bad">{--LOGOUT_FAILED--}</span>'; break;
		case getCode('DATA_INVALID')       : $message = '{--MAIL_DATA_INVALID--}'; break;
		case getCode('POSSIBLE_INVALID')   : $message = '{--MAIL_POSSIBLE_INVALID--}'; break;
		case getCode('USER_404')           : $message = '{--USER_404--}'; break;
		case getCode('STATS_404')          : $message = '{--MAIL_STATS_404--}'; break;
		case getCode('ALREADY_CONFIRMED')  : $message = '{--MAIL_ALREADY_CONFIRMED--}'; break;
		case getCode('WRONG_PASS')         : $message = '{--LOGIN_WRONG_PASS--}'; break;
		case getCode('WRONG_ID')           : $message = '{--LOGIN_WRONG_ID--}'; break;
		case getCode('ACCOUNT_LOCKED')     : $message = '{--LOGIN_STATUS_LOCKED--}'; break;
		case getCode('ACCOUNT_UNCONFIRMED'): $message = '{--LOGIN_STATUS_UNCONFIRMED--}'; break;
		case getCode('COOKIES_DISABLED')   : $message = '{--LOGIN_COOKIES_DISABLED--}'; break;
		case getCode('BEG_SAME_AS_OWN')    : $message = '{--BEG_SAME_USERID_AS_OWN--}'; break;
		case getCode('LOGIN_FAILED')       : $message = '{--GUEST_LOGIN_FAILED_GENERAL--}'; break;
		case getCode('MODULE_MEMBER_ONLY') : $message = '{%message,MODULE_MEMBER_ONLY=' . getRequestElement('mod') . '%}'; break;
		case getCode('OVERLENGTH')         : $message = '{--MEMBER_TEXT_OVERLENGTH--}'; break;
		case getCode('URL_FOUND')          : $message = '{--MEMBER_TEXT_CONTAINS_URL--}'; break;
		case getCode('SUBJECT_URL')        : $message = '{--MEMBER_SUBJECT_CONTAINS_URL--}'; break;
		case getCode('BLIST_URL')          : $message = '{--MEMBER_URL_BLACK_LISTED--}<br />{--MEMBER_BLIST_TIME--}: ' . generateDateTime(getRequestElement('blist'), 0); break;
		case getCode('NO_RECS_LEFT')       : $message = '{--MEMBER_SELECTED_MORE_RECS--}'; break;
		case getCode('INVALID_TAGS')       : $message = '{--MEMBER_HTML_INVALID_TAGS--}'; break;
		case getCode('MORE_POINTS')        : $message = '{--MEMBER_MORE_POINTS_NEEDED--}'; break;
		case getCode('MORE_RECEIVERS1')    : $message = '{--MEMBER_ENTER_MORE_RECEIVERS--}'; break;
		case getCode('MORE_RECEIVERS2')    : $message = '{--MEMBER_NO_MORE_RECEIVERS_FOUND--}'; break;
		case getCode('MORE_RECEIVERS3')    : $message = '{--MEMBER_ENTER_MORE_MIN_RECEIVERS--}'; break;
		case getCode('INVALID_URL')        : $message = '{--MEMBER_ENTER_INVALID_URL--}'; break;
		case getCode('NO_MAIL_TYPE')       : $message = '{--MEMBER_NO_MAIL_TYPE_SELECTED--}'; break;
		case getCode('UNKNOWN_ERROR')      : $message = '{--LOGIN_UNKNOWN_ERROR--}'; break;
		case getCode('UNKNOWN_STATUS')     : $message = '{--LOGIN_UNKNOWN_STATUS--}'; break;
		case getCode('PROFILE_UPDATED')    : $message = '{--MEMBER_PROFILE_UPDATED--}'; break;

		case getCode('ERROR_MAILID'):
			if (isExtensionActive('mailid', true)) {
				$message = '{--ERROR_CONFIRMING_MAIL--}';
			} else {
				$message = '{%pipe,generateExtensionInactiveNotInstalledMessage=mailid%}';
			}
			break;

		case getCode('EXTENSION_PROBLEM'):
			if (isGetRequestElementSet('ext')) {
				$message = '{%pipe,generateExtensionInactiveNotInstalledMessage=' . getRequestElement('ext') . '%}';
			} else {
				$message = '{--EXTENSION_PROBLEM_UNSET_EXT--}';
			}
			break;

		case getCode('URL_TIME_LOCK'):
			// @TODO Move this SQL code into a function, let's say 'getTimestampFromPoolId($id) ?
			$result = SQL_QUERY_ESC("SELECT `timestamp` FROM `{?_MYSQL_PREFIX?}_pool` WHERE `id`=%s LIMIT 1",
				array(bigintval(getRequestElement('id'))), __FUNCTION__, __LINE__);

			// Load timestamp from last order
			$content = SQL_FETCHARRAY($result);

			// Free memory
			SQL_FREERESULT($result);

			// Translate it for templates
			$content['timestamp'] = generateDateTime($content['timestamp'], 1);

			// Calculate hours...
			$content['hours'] = round(getUrlTlock() / 60 / 60);

			// Minutes...
			$content['minutes'] = round((getUrlTlock() - $content['hours'] * 60 * 60) / 60);

			// And seconds
			$content['seconds'] = round(getUrlTlock() - $content['hours'] * 60 * 60 - $content['minutes'] * 60);

			// Finally contruct the message
			$message = loadTemplate('tlock_message', true, $content);
			break;

		default:
			// Missing/invalid code
			$message = '{%message,UNKNOWN_MAILID_CODE=' . $code . '%}';

			// Log it
			logDebugMessage(__FUNCTION__, __LINE__, $message);
			break;
	} // END - switch

	// Return the message
	return $message;
}

// Function taken from user comments on www.php.net / function isInStringIgnoreCase()
function isUrlValidSimple ($url) {
	// Prepare URL
	$url = secureString(str_replace("\\", '', compileRawCode(urldecode($url))));

	// Allows http and https
	$http      = "(http|https)+(:\/\/)";
	// Test domain
	$domain1   = "([[:alnum:]]([-[:alnum:]])*\.)?([[:alnum:]][-[:alnum:]\.]*[[:alnum:]])(\.[[:alpha:]]{2,5})?";
	// Test double-domains (e.g. .de.vu)
	$domain2   = "([-[:alnum:]])?(\.[[:alnum:]][-[:alnum:]\.]*[[:alnum:]])(\.[[:alpha:]]{2,5})(\.[[:alpha:]]{2,5})?";
	// Test IP number
	$ip        = "([[:digit:]]{1,3})\.([[:digit:]]{1,3})\.([[:digit:]]{1,3})\.([[:digit:]]{1,3})";
	// ... directory
	$dir       = "((/)+([-_\.[:alnum:]])+)*";
	// ... page
	$page      = "/([-_[:alnum:]][-\._[:alnum:]]*\.[[:alnum:]]{2,5})?";
	// ... and the string after and including question character
	$getstring1 = "([\?/]([[:alnum:]][-\._%[:alnum:]]*(=)?([-\@\._:%[:alnum:]])+)(&([[:alnum:]]([-_%[:alnum:]])*(=)?([-\@\[\._:%[:alnum:]])+(\])*))*)?";
	// Pattern for URLs like http://url/dir/doc.html?var=value
	$pattern['d1dpg1']  = $http . $domain1 . $dir . $page . $getstring1;
	$pattern['d2dpg1']  = $http . $domain2 . $dir . $page . $getstring1;
	$pattern['ipdpg1']  = $http . $ip . $dir . $page . $getstring1;
	// Pattern for URLs like http://url/dir/?var=value
	$pattern['d1dg1']  = $http . $domain1 . $dir.'/' . $getstring1;
	$pattern['d2dg1']  = $http . $domain2 . $dir.'/' . $getstring1;
	$pattern['ipdg1']  = $http . $ip . $dir.'/' . $getstring1;
	// Pattern for URLs like http://url/dir/page.ext
	$pattern['d1dp']  = $http . $domain1 . $dir . $page;
	$pattern['d1dp']  = $http . $domain2 . $dir . $page;
	$pattern['ipdp']  = $http . $ip . $dir . $page;
	// Pattern for URLs like http://url/dir
	$pattern['d1d']  = $http . $domain1 . $dir;
	$pattern['d2d']  = $http . $domain2 . $dir;
	$pattern['ipd']  = $http . $ip . $dir;
	// Pattern for URLs like http://url/?var=value
	$pattern['d1g1']  = $http . $domain1 . '/' . $getstring1;
	$pattern['d2g1']  = $http . $domain2 . '/' . $getstring1;
	$pattern['ipg1']  = $http . $ip . '/' . $getstring1;
	// Pattern for URLs like http://url?var=value
	$pattern['d1g12']  = $http . $domain1 . $getstring1;
	$pattern['d2g12']  = $http . $domain2 . $getstring1;
	$pattern['ipg12']  = $http . $ip . $getstring1;

	// Test all patterns
	$reg = false;
	foreach ($pattern as $key => $pat) {
		// Debug regex?
		if (isDebugRegularExpressionEnabled()) {
			// @TODO Are these convertions still required?
			$pat = str_replace('.', '&#92;&#46;', $pat);
			$pat = str_replace('@', '&#92;&#64;', $pat);
			//* DEBUG: */ debugOutput($key . '=&nbsp;' . $pat);
		} // END - if

		// Check if expression matches
		$reg = ($reg || preg_match(('^' . $pat . '^'), $url));

		// Does it match?
		if ($reg === true) {
			break;
		} // END - if
	} // END - foreach

	// Return true/false
	return $reg;
}

// Wtites data to a config.php-style file
// @TODO Rewrite this function to use readFromFile() and writeToFile()
function changeDataInFile ($FQFN, $comment, $prefix, $suffix, $inserted, $seek=0) {
	// Initialize some variables
	$done = false;
	$seek++;
	$next  = -1;
	$found = false;

	// Is the file there and read-/write-able?
	if ((isFileReadable($FQFN)) && (is_writeable($FQFN))) {
		$search = 'CFG: ' . $comment;
		$tmp = $FQFN . '.tmp';

		// Open the source file
		$fp = fopen($FQFN, 'r') or debug_report_bug(__FUNCTION__, __LINE__, 'Cannot read. file=' . basename($FQFN));

		// Is the resource valid?
		if (is_resource($fp)) {
			// Open temporary file
			$fp_tmp = fopen($tmp, 'w') or debug_report_bug(__FUNCTION__, __LINE__, 'Cannot write. tmp=' . basename($tmp) . ',file=' . $FQFN);

			// Is the resource again valid?
			if (is_resource($fp_tmp)) {
				// Mark temporary file as readable
				$GLOBALS['file_readable'][$tmp] = true;

				// Start reading
				while (!feof($fp)) {
					// Read from source file
					$line = fgets ($fp, 1024);

					if (isInString($search, $line)) {
						$next = '0';
						$found = true;
					} // END - if

					if ($next > -1) {
						if ($next === $seek) {
							$next = -1;
							$line = $prefix . $inserted . $suffix . "\n";
						} else {
							$next++;
						}
					} // END - if

					// Write to temp file
					fwrite($fp_tmp, $line);
				} // END - while

				// Close temp file
				fclose($fp_tmp);

				// Finished writing tmp file
				$done = true;
			} // END - if

			// Close source file
			fclose($fp);

			if (($done === true) && ($found === true)) {
				// Copy back tmp file and delete tmp :-)
				copyFileVerified($tmp, $FQFN, 0644);
				return removeFile($tmp);
			} elseif ($found === false) {
				outputHtml('<strong>CHANGE:</strong> 404!');
			} else {
				outputHtml('<strong>TMP:</strong> UNDONE!');
			}
		}
	} else {
		// File not found, not readable or writeable
		debug_report_bug(__FUNCTION__, __LINE__, 'File not readable/writeable. file=' . basename($FQFN));
	}

	// An error was detected!
	return false;
}

// Send notification to admin
function sendAdminNotification ($subject, $templateName, $content = array(), $userid = NULL) {
	if ((isExtensionInstalledAndNewer('admins', '0.4.1')) && (function_exists('sendAdminsEmails'))) {
		// Send new way
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'admins=Y,subject=' . $subject . ',templateName=' . $templateName . ' - OKAY!');
		sendAdminsEmails($subject, $templateName, $content, $userid);
	} else {
		// Send out-dated way
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'admins=N,subject=' . $subject . ',templateName=' . $templateName . ' - OUT-DATED!');
		$message = loadEmailTemplate($templateName, $content, $userid);
		sendAdminEmails($subject, $message);
	}
}

// Debug message logger
function logDebugMessage ($funcFile, $line, $message, $force=true) {
	// Is debug mode enabled?
	if ((isDebugModeEnabled()) || ($force === true)) {
		// Remove CRLF
		$message = str_replace("\r", '', str_replace("\n", '', $message));

		// Log this message away
		appendLineToFile(getPath() . getCachePath() . 'debug.log', generateDateTime(time(), '4') . '|' . getModule(false) . '|' . basename($funcFile) . '|' . $line . '|' . $message);
	} // END - if
}

// Handle extra values
function handleExtraValues ($filterFunction, $value, $extraValue) {
	// Default is the value itself
	$ret = $value;

	// Do we have a special filter function?
	if (!empty($filterFunction)) {
		// Does the filter function exist?
		if (function_exists($filterFunction)) {
			// Do we have extra parameters here?
			if (!empty($extraValue)) {
				// Put both parameters in one new array by default
				$args = array($value, $extraValue);

				// If we have an array simply use it and pre-extend it with our value
				if (is_array($extraValue)) {
					// Make the new args array
					$args = merge_array(array($value), $extraValue);
				} // END - if

				// Call the multi-parameter call-back
				$ret = call_user_func_array($filterFunction, $args);
			} else {
				// One parameter call
				$ret = call_user_func($filterFunction, $value);
			}
		} // END - if
	} // END - if

	// Return the value
	return $ret;
}

// Converts timestamp selections into a timestamp
function convertSelectionsToEpocheTime (array &$postData, array &$content, &$id, &$skip) {
	// Init test variable
	$skip  = false;
	$test2 = '';

	// Get last three chars
	$test = substr($id, -3);

	// Improved way of checking! :-)
	if (in_array($test, array('_ye', '_mo', '_we', '_da', '_ho', '_mi', '_se'))) {
		// Found a multi-selection for timings?
		$test = substr($id, 0, -3);
		if ((isset($postData[$test . '_ye'])) && (isset($postData[$test . '_mo'])) && (isset($postData[$test . '_we'])) && (isset($postData[$test . '_da'])) && (isset($postData[$test . '_ho'])) && (isset($postData[$test . '_mi'])) && (isset($postData[$test . '_se'])) && ($test != $test2)) {
			// Generate timestamp
			$postData[$test] = createEpocheTimeFromSelections($test, $postData);
			$content[] = sprintf("`%s`='%s'", $test, $postData[$test]);
			$GLOBALS['skip_config'][$test] = true;

			// Remove data from array
			foreach (array('ye', 'mo', 'we', 'da', 'ho', 'mi', 'se') as $rem) {
				unset($postData[$test . '_' . $rem]);
			} // END - foreach

			// Skip adding
			unset($id);
			$skip = true;
			$test2 = $test;
		} // END - if
	} // END - if
}

// Reverts the german decimal comma into Computer decimal dot
function convertCommaToDot ($str) {
	// Default float is not a float... ;-)
	$float = false;

	// Which language is selected?
	switch (getLanguage()) {
		case 'de': // German language
			// Remove german thousand dots first
			$str = str_replace('.', '', $str);

			// Replace german commata with decimal dot and cast it
			$float = (float)str_replace(',', '.', $str);
			break;

		default: // US and so on
			// Remove thousand dots first and cast
			$float = (float)str_replace(',', '', $str);
			break;
	}

	// Return float
	return $float;
}

// Handle menu-depending failed logins and return the rendered content
function handleLoginFailures ($accessLevel) {
	// Default output is empty ;-)
	$OUT = '';

	// Is the session data set?
	if ((isSessionVariableSet('mailer_' . $accessLevel . '_failures')) && (isSessionVariableSet('mailer_' . $accessLevel . '_last_failure'))) {
		// Ignore zero values
		if (getSession('mailer_' . $accessLevel . '_failures') > 0) {
			// Non-guest has login failures found, get both data and prepare it for template
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'accessLevel=' . $accessLevel . '<br />');
			$content = array(
				'login_failures' => 'mailer_' . $accessLevel . '_failures',
				'last_failure'   => generateDateTime(getSession('mailer_' . $accessLevel . '_last_failure'), 2)
			);

			// Load template
			$OUT = loadTemplate('login_failures', true, $content);
		} // END - if

		// Reset session data
		setSession('mailer_' . $accessLevel . '_failures', '');
		setSession('mailer_' . $accessLevel . '_last_failure', '');
	} // END - if

	// Return rendered content
	return $OUT;
}

// Rebuild cache
function rebuildCache ($cache, $inc = '', $force = false) {
	// Debug message
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("cache=%s, inc=%s, force=%s", $cache, $inc, intval($force)));

	// Shall I remove the cache file?
	if (isCacheInstanceValid()) {
		// Rebuild cache
		if ($GLOBALS['cache_instance']->loadCacheFile($cache)) {
			// Destroy it
			$GLOBALS['cache_instance']->removeCacheFile($force);
		} // END - if

		// Include file given?
		if (!empty($inc)) {
			// Construct FQFN
			$inc = sprintf("inc/loader/load-%s.php", $inc);

			// Is the include there?
			if (isIncludeReadable($inc)) {
				// And rebuild it from scratch
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'inc=' . $inc . ' - LOADED!');
				loadInclude($inc);
			} else {
				// Include not found
				logDebugMessage(__FUNCTION__, __LINE__, 'Include ' . $inc . ' not found. cache=' . $cache);
			}
		} // END - if
	} // END - if
}

// Determines the real remote address
function determineRealRemoteAddress ($remoteAddr = false) {
	// Is a proxy in use?
	if ((isset($_SERVER['HTTP_X_FORWARDED_FOR'])) && (!$remoteAddr)) {
		// Proxy was used
		$address = $_SERVER['HTTP_X_FORWARDED_FOR'];
	} elseif ((isset($_SERVER['HTTP_CLIENT_IP'])) && (!$remoteAddr)) {
		// Yet, another proxy
		$address = $_SERVER['HTTP_CLIENT_IP'];
	} else {
		// The regular address when no proxy was used
		$address = $_SERVER['REMOTE_ADDR'];
	}

	// This strips out the real address from proxy output
	if (strstr($address, ',')) {
		$addressArray = explode(',', $address);
		$address = $addressArray[0];
	} // END - if

	// Return the result
	return $address;
}

// Adds a bonus mail to the queue
// This is a high-level function!
function addNewBonusMail ($data, $mode = '', $output = true) {
	// Use mode from data if not set and availble ;-)
	if ((empty($mode)) && (isset($data['mail_mode']))) {
		$mode = $data['mail_mode'];
	} // END - if

	// Generate receiver list
	$receiver = generateReceiverList($data['cat'], $data['receiver'], $mode);

	// Receivers added?
	if (!empty($receiver)) {
		// Add bonus mail to queue
		addBonusMailToQueue(
			$data['subject'],
			$data['text'],
			$receiver,
			$data['points'],
			$data['seconds'],
			$data['url'],
			$data['cat'],
			$mode,
			$data['receiver']
		);

		// Mail inserted into bonus pool
		if ($output === true) {
			displayMessage('{--ADMIN_BONUS_SEND--}');
		} // END - if
	} elseif ($output === true) {
		// More entered than can be reached!
		displayMessage('{--ADMIN_MORE_SELECTED--}');
	} else {
		// Debug log
		logDebugMessage(__FUNCTION__, __LINE__, 'cat=' . $data['cat'] . ',receiver=' . $data['receiver'] . ',data=' . base64_encode(serialize($data)) . ' More selected, than available!');
	}
}

// Determines referral id and sets it
function determineReferralId () {
	// Skip this in non-html-mode and outside ref.php
	if ((!isHtmlOutputMode()) && (basename($_SERVER['PHP_SELF']) != 'ref.php')) {
		return false;
	} // END - if

	// Check if refid is set
	if (isReferralIdValid()) {
		// This is fine...
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using refid from GLOBALS (' . getReferralId() . ')');
	} elseif (isPostRequestElementSet('refid')) {
		// Get referral id from POST element refid
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using refid from POST data (' . postRequestElement('refid') . ')');
		setReferralId(secureString(postRequestElement('refid')));
	} elseif (isGetRequestElementSet('refid')) {
		// Get referral id from GET parameter refid
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using refid from GET data (' . getRequestElement('refid') . ')');
		setReferralId(secureString(getRequestElement('refid')));
	} elseif (isGetRequestElementSet('ref')) {
		// Set refid=ref (the referral link uses such variable)
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using ref from GET data (' . getRequestElement('refid') . ')');
		setReferralId(secureString(getRequestElement('ref')));
	} elseif ((isGetRequestElementSet('user')) && (basename($_SERVER['PHP_SELF']) == 'click.php')) {
		// The variable user comes from  click.php
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using user from GET data (' . getRequestElement('user') . ')');
		setReferralId(bigintval(getRequestElement('user')));
	} elseif ((isSessionVariableSet('refid')) && (isValidUserId(getSession('refid')))) {
		// Set session refid as global
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using refid from SESSION data (' . getSession('refid') . ')');
		setReferralId(bigintval(getSession('refid')));
	} elseif (isRandomReferralIdEnabled()) {
		// Select a random user which has confirmed enougth mails
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Checking random referral id');
		setReferralId(determineRandomReferralId());
	} elseif ((isExtensionInstalledAndNewer('sql_patches', '0.1.2')) && (isValidUserId(getDefRefid()))) {
		// Set default refid as refid in URL
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using default refid (' . getDefRefid() . ')');
		setReferralId(getDefRefid());
	} else {
		// No default id when sql_patches is not installed or none set
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using NULL as refid');
		setReferralId(NULL);
	}

	// Set cookie when default refid > 0
	if ((!isSessionVariableSet('refid')) || (!isValidUserId(getReferralId())) || ((!isValidUserId(getSession('refid'))) && (isExtensionInstalledAndNewer('sql_patches', '0.1.2')) && (isValidUserId(getDefRefid())))) {
		// Default is not found
		$found = false;

		// Do we have nickname or userid set?
		if ((isExtensionActive('nickname')) && (isNicknameUsed(getReferralId()))) {
			// Nickname in URL, so load the id
			$found = fetchUserData(getReferralId(), 'nickname');

			// If we found it, use the userid as referral id
			if ($found === true) {
				// Set the userid as 'refid'
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using refid from user account by nickname (' . getUserData('userid') . ')');
				setReferralId(getUserData('userid'));
			} // END - if
		} elseif (isValidUserId(getReferralId())) {
			// Direct userid entered
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using direct userid (' . getReferralId() . ')');
			$found = fetchUserData(getReferralId());
		}

		// Is the record valid?
		if ((($found === false) || (!isUserDataValid())) && (isExtensionInstalledAndNewer('sql_patches', '0.1.2'))) {
			// No, then reset referral id
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using default refid (' . getDefRefid() . ')');
			setReferralId(getDefRefid());
		} // END - if

		// Set cookie
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Saving refid to session (' . getReferralId() . ') #1');
		setSession('refid', getReferralId());
	} elseif ((!isReferralIdValid()) || (!fetchUserData(getReferralId()))) {
		// Not valid!
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Not valid referral id (' . getReferralId() . '), setting NULL in session');
		setReferralId(NULL);
		setSession('refid', NULL);
	} else {
		// Set it from GLOBALS array in session
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Saving refid to session (' . getReferralId() . ') #2');
		setSession('refid', getReferralId());
	}

	// Return determined refid
	return getReferralId();
}

// Enables the reset mode and runs it
function doReset () {
	// Enable the reset mode
	$GLOBALS['reset_enabled'] = true;

	// Run filters
	runFilterChain('reset');
}

// Enables the reset mode (hourly, weekly and monthly) and runs it
function doHourly () {
	// Enable the hourly reset mode
	$GLOBALS['hourly_enabled'] = true;

	// Run filters (one always!)
	runFilterChain('hourly');
}

// Our shutdown-function
function shutdown () {
	// Call the filter chain 'shutdown'
	runFilterChain('shutdown', null);

	// Check if not in installation phase and the link is up
	if ((!isInstallationPhase()) && (SQL_IS_LINK_UP())) {
		// Close link
		SQL_CLOSE(__FUNCTION__, __LINE__);
	} elseif (!isInstallationPhase()) {
		// No database link
		debug_report_bug(__FUNCTION__, __LINE__, 'Database link is already down, while shutdown is running.');
	}

	// Stop executing here
	exit;
}

// Init member id
function initMemberId () {
	$GLOBALS['member_id'] = '0';
}

// Setter for member id
function setMemberId ($memberid) {
	// We should not set member id to zero
	if ($memberid == '0') {
		debug_report_bug(__FUNCTION__, __LINE__, 'Userid should not be set zero.');
	} // END - if

	// Set it secured
	$GLOBALS['member_id'] = bigintval($memberid);
}

// Getter for member id or returns zero
function getMemberId () {
	// Default member id
	$memberid = '0';

	// Is the member id set?
	if (isMemberIdSet()) {
		// Then use it
		$memberid = $GLOBALS['member_id'];
	} // END - if

	// Return it
	return $memberid;
}

// Checks ether the member id is set
function isMemberIdSet () {
	return (isset($GLOBALS['member_id']));
}

// Setter for extra title
function setExtraTitle ($extraTitle) {
	$GLOBALS['extra_title'] = $extraTitle;
}

// Getter for extra title
function getExtraTitle () {
	// Is the extra title set?
	if (!isExtraTitleSet()) {
		// No, then abort here
		debug_report_bug(__FUNCTION__, __LINE__, 'extra_title is not set!');
	} // END - if

	// Return it
	return $GLOBALS['extra_title'];
}

// Checks if the extra title is set
function isExtraTitleSet () {
	return ((isset($GLOBALS['extra_title'])) && (!empty($GLOBALS['extra_title'])));
}

/**
 * Reads a directory recursively by default and searches for files not matching
 * an exclusion pattern. You can now keep the exclusion pattern empty for reading
 * a whole directory.
 *
 * @param	$baseDir			Relative base directory to PATH to scan from
 * @param	$prefix				Prefix for all positive matches (which files should be found)
 * @param	$fileIncludeDirs	Wether to include directories in the final output array
 * @param	$addBaseDir			Wether to add $baseDir to all array entries
 * @param	$excludeArray		Excluded files and directories, these must be full files names, e.g. 'what-' will exclude all files named 'what-' but won't exclude 'what-foo.php'
 * @param	$extension			File extension for all positive matches
 * @param	$excludePattern		Regular expression to exclude more files (preg_match())
 * @param	$recursive			Wether to scan recursively
 * @param	$suffix				Suffix for positive matches ($extension will be appended, too)
 * @return	$foundMatches		All found positive matches for above criteria
 */
function getArrayFromDirectory ($baseDir, $prefix, $fileIncludeDirs = false, $addBaseDir = true, $excludeArray = array(), $extension = '.php', $excludePattern = '@(\.|\.\.)$@', $recursive = true, $suffix = '') {
	// Add default entries we should always exclude
	$excludeArray[] = '.';         // Current directory
	$excludeArray[] = '..';        // Parent directory
	$excludeArray[] = '.svn';      // Directories created by Subversion
	$excludeArray[] = '.htaccess'; // Directory protection files (mostly)

	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'baseDir=' . $baseDir . ',prefix=' . $prefix . ' - Entered!');
	// Init found includes
	$foundMatches = array();

	// Open directory
	$dirPointer = opendir(getPath() . $baseDir) or debug_report_bug(__FUNCTION__, __LINE__, 'Cannot read directory ' . basename($baseDir) . '.');

	// Read all entries
	while ($baseFile = readdir($dirPointer)) {
		// Exclude '.', '..' and entries in $excludeArray automatically
		if (in_array($baseFile, $excludeArray, true))  {
			// Exclude them
			//* DEBUG: */ debugOutput('excluded=' . $baseFile);
			continue;
		} // END - if

		// Construct include filename and FQFN
		$fileName = $baseDir . $baseFile;
		$FQFN = getPath() . $fileName;

		// Remove double slashes
		$FQFN = str_replace('//', '/', $FQFN);

		// Check if the base filenname matches an exclusion pattern and if the pattern is not empty
		if ((!empty($excludePattern)) && (preg_match($excludePattern, $baseFile, $match))) {
			// Debug message
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'baseDir=' . $baseDir . ',baseFile=' . $baseFile . ',FQFN=' . $FQFN);

			// Exclude this one
			continue;
		} // END - if

		// Skip also files with non-matching prefix genericly
		if (($recursive === true) && (isDirectory($FQFN))) {
			// Is a redirectory so read it as well
			$foundMatches = merge_array($foundMatches, getArrayFromDirectory($baseDir . $baseFile . '/', $prefix, $fileIncludeDirs, $addBaseDir, $excludeArray, $extension, $excludePattern, $recursive));

			// And skip further processing
			continue;
		} elseif (!isFilePrefixFound($baseFile, $prefix)) {
			// Skip this file
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Invalid prefix in file ' . $baseFile . ', prefix=' . $prefix);
			continue;
		} elseif ((!empty($suffix)) && (substr($baseFile, -(strlen($suffix . $extension)), (strlen($suffix . $extension))) != $suffix . $extension)) {
			// Skip wrong suffix as well
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Invalid suffix in file ' . $baseFile . ', suffix=' . $suffix);
			continue;
		} elseif (!isFileReadable($FQFN)) {
			// Not readable so skip it
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'File ' . $FQFN . ' is not readable!');
			continue;
		}

		// Get file' extension (last 4 chars)
		$fileExtension = substr($baseFile, -4, 4);

		// Is the file a PHP script or other?
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'baseDir=' . $baseDir . ',prefix=' . $prefix . ',baseFile=' . $baseFile);
		if (($fileExtension == '.php') || (($fileIncludeDirs === true) && (isDirectory($FQFN)))) {
			// Is this a valid include file?
			if ($extension == '.php') {
				// Remove both for extension name
				$extName = substr($baseFile, strlen($prefix), -4);

				// Add file with or without base path
				if ($addBaseDir === true) {
					// With base path
					$foundMatches[] = $fileName;
				} else {
					// No base path
					$foundMatches[] = $baseFile;
				}
			} else {
				// We found .php file but should not search for them, why?
				debug_report_bug(__FUNCTION__, __LINE__, 'We should find files with extension=' . $extension . ', but we found a PHP script. (baseFile=' . $baseFile . ')');
			}
		} elseif ($fileExtension == $extension) {
			// Other, generic file found
			$foundMatches[] = $fileName;
		}
	} // END - while

	// Close directory
	closedir($dirPointer);

	// Sort array
	sort($foundMatches);

	// Return array with include files
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, '- Left!');
	return $foundMatches;
}

// Checks wether $prefix is found in $fileName
function isFilePrefixFound ($fileName, $prefix) {
	// @TODO Find a way to cache this
	return (substr($fileName, 0, strlen($prefix)) == $prefix);
}

// Maps a module name into a database table name
function mapModuleToTable ($moduleName) {
	// Map only these, still lame code...
	switch ($moduleName) {
		// 'index' is the guest's menu
		case 'index': $moduleName = 'guest';  break;
		// ... and 'login' the member's menu
		case 'login': $moduleName = 'member'; break;
		// Anything else will not be mapped, silently.
	} // END - switch

	// Return result
	return $moduleName;
}

// Add SQL debug data to array for later output
function addSqlToDebug ($result, $sqlString, $timing, $F, $L) {
	// Do we have cache?
	if (!isset($GLOBALS['debug_sql_available'])) {
		// Check it and cache it in $GLOBALS
		$GLOBALS['debug_sql_available'] = ((isConfigurationLoaded()) && (isDisplayDebugSqlEnabled()));
	} // END - if
	
	// Don't execute anything here if we don't need or ext-other is missing
	if ($GLOBALS['debug_sql_available'] === false) {
		return;
	} // END - if

	// Already executed?
	if (isset($GLOBALS['debug_sqls'][$F][$L][$sqlString])) {
		// Then abort here, we don't need to profile a query twice
		return;
	} // END - if

	// Remeber this as profiled (or not, but we don't care here)
	$GLOBALS['debug_sqls'][$F][$L][$sqlString] = true;

	// Generate record
	$record = array(
		'num_rows' => SQL_NUMROWS($result),
		'affected' => SQL_AFFECTEDROWS(),
		'sql_str'  => $sqlString,
		'timing'   => $timing,
		'file'     => basename($F),
		'line'     => $L
	);

	// Add it
	$GLOBALS['debug_sqls'][] = $record;
}

// Initializes the cache instance
function initCacheInstance () {
	// Check for double-initialization
	if (isset($GLOBALS['cache_instance'])) {
		// This should not happen and must be fixed
		debug_report_bug(__FUNCTION__, __LINE__, 'Double initialization of cache system detected. cache_instance[]=' . gettype($GLOBALS['cache_instance']));
	} // END - if

	// Load include for CacheSystem class
	loadIncludeOnce('inc/classes/cachesystem.class.php');

	// Initialize cache system only when it's needed
	$GLOBALS['cache_instance'] = new CacheSystem();

	// Did it work?
	if ($GLOBALS['cache_instance']->getStatusCode() != 'done') {
		// Failed to initialize cache sustem
		debug_report_bug(__FUNCTION__, __LINE__, 'Cache system returned with unexpected error. getStatusCode()=' . $GLOBALS['cache_instance']->getStatusCode());
	} // END - if
}

// Getter for message from array or raw message
function getMessageFromIndexedArray ($message, $pos, $array) {
	// Check if the requested message was found in array
	if (isset($array[$pos])) {
		// ... if yes then use it!
		$ret = $array[$pos];
	} else {
		// ... else use default message
		$ret = $message;
	}

	// Return result
	return $ret;
}

// Convert ';' to ', ' for e.g. receiver list
function convertReceivers ($old) {
	return str_replace(';', ', ', $old);
}

// Get a module from filename and access level
function getModuleFromFileName ($file, $accessLevel) {
	// Default is 'invalid';
	$modCheck = 'invalid';

	// @TODO This is still very static, rewrite it somehow
	switch ($accessLevel) {
		case 'admin':
			$modCheck = 'admin';
			break;

		case 'sponsor':
		case 'guest':
		case 'member':
			$modCheck = getModule();
			break;

		default: // Unsupported file name / access level
			debug_report_bug(__FUNCTION__, __LINE__, 'Unsupported file name=' . basename($file) . '/access level=' . $accessLevel);
			break;
	}

	// Return result
	return $modCheck;
}

// Encodes an URL for adding session id, etc.
function encodeUrl ($url, $outputMode = '0') {
	// Do we have already have a PHPSESSID inside or view.php is called? Then abort here
	if ((isInStringIgnoreCase(session_name(), $url)) || (isRawOutputMode())) {
		// Raw output mode detected or session_name() found in URL
		return $url;
	} // END - if

	// Do we have a valid session?
	if (((!isset($GLOBALS['valid_session'])) || ($GLOBALS['valid_session'] === false) || (!isset($_COOKIE[session_name()]))) && (isSpider() === false)) {
		// Invalid session
		// Determine right separator
		$separator = '&amp;';
		if (!isInString('?', $url)) {
			// No question mark
			$separator = '?';
		} elseif ((!isHtmlOutputMode()) || ($outputMode != '0')) {
			// Non-HTML mode (or forced non-HTML mode
			$separator = '&';
		}

		// Add it to URL
		if (session_id() != '') {
			$url .= $separator . session_name() . '=' . session_id();
		} // END - if
	} // END - if

	// Add {?URL?} ?
	if ((substr($url, 0, strlen(getUrl())) != getUrl()) && (substr($url, 0, 7) != '{?URL?}') && (substr($url, 0, 7) != 'http://') && (substr($url, 0, 8) != 'https://')) {
		// Add it
		$url = '{?URL?}/' . $url;
	} // END - if

	// Return the URL
	return $url;
}

// Simple check for spider
function isSpider () {
	// Get the UA and trim it down
	$userAgent = trim(detectUserAgent(true));

	// It should not be empty, if so it is better a spider/bot
	if (empty($userAgent)) {
		// It is a spider/bot
		return true;
	} // END - if

	// Is it a spider?
	return ((isInStringIgnoreCase('spider', $userAgent)) || (isInStringIgnoreCase('slurp', $userAgent)) || (isInStringIgnoreCase('bot', $userAgent)) || (isInStringIgnoreCase('archiver', $userAgent)));
}

// Function to search for the last modified file
function searchDirsRecursive ($dir, &$last_changed, $lookFor = 'Date') {
	// Get dir as array
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'dir=' . $dir);
	// Does it match what we are looking for? (We skip a lot files already!)
	// RegexPattern to exclude  ., .., .revision,  .svn, debug.log or .cache in the filenames
	$excludePattern = '@(\.revision|\.svn|debug\.log|\.cache|config\.php)$@';

	$ds = getArrayFromDirectory($dir, '', false, true, array(), '.php', $excludePattern);
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'count(ds)='.count($ds));

	// Walk through all entries
	foreach ($ds as $d) {
		// Generate proper FQFN
		$FQFN = str_replace('//', '/', getPath() . $dir . '/' . $d);

		// Is it a file and readable?
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'dir=' . $dir . ',d=' . $d);
		if (isFileReadable($FQFN)) {
			// $FQFN is a readable file so extract the requested data from it
			$check = extractRevisionInfoFromFile($FQFN, $lookFor);
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'File: ' . $d . ' found. check=' . $check);

			// Is the file more recent?
			if ((!isset($last_changed[$lookFor])) || ($last_changed[$lookFor] < $check)) {
				// This file is newer as the file before
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'NEWER!');
				$last_changed['path_name'] = $FQFN;
				$last_changed[$lookFor] = $check;
			} // END - if
		} else {
			// Not readable
			/* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'File: ' . $d . ' not readable or directory.');
		}
	} // END - foreach
}

// Handles the braces [] of a field (e.g. value of 'name' attribute)
function handleFieldWithBraces ($field) {
	// Are there braces [] at the end?
	if (substr($field, -2, 2) == '[]') {
		// Try to find one and replace it. I do it this way to allow easy
		// extending of this code.
		foreach (array('admin_list_builder_id_value') as $key) {
			// Is the cache entry set?
			if (isset($GLOBALS[$key])) {
				// Insert it
				$field = str_replace('[]', '[' . $GLOBALS[$key] . ']', $field);

				// And abort
				break;
			} // END - if
		} // END - foreach
	} // END - if

	// Return it
	return $field;
}

// Converts a zero or NULL to word 'NULL'
function convertZeroToNull ($number) {
	// Is it a valid username?
	if ((!is_null($number)) && (!empty($number)) && ($number > 0)) {
		// Always secure it
		$number = bigintval($number);
	} else {
		// Is not valid or zero
		$number = 'NULL';
	}

	// Return it
	return $number;
}

// Converts a NULL to zero
function convertNullToZero ($number) {
	// Is it a valid username?
	if ((!is_null($number)) && (!empty($number)) && ($number > 0)) {
		// Always secure it
		$number = bigintval($number);
	} else {
		// Is not valid or zero
		$number = '0';
	}

	// Return it
	return $number;
}

// Capitalizes a string with underscores, e.g.: some_foo_string will become SomeFooString
// Note: This function is cached
function capitalizeUnderscoreString ($str) {
	// Do we have cache?
	if (!isset($GLOBALS[__FUNCTION__][$str])) {
		// Init target string
		$capitalized = '';

		// Explode it with the underscore, but rewrite dashes to underscore before
		$strArray = explode('_', str_replace('-', '_', $str));

		// "Walk" through all elements and make them lower-case but first upper-case
		foreach ($strArray as $part) {
			// Capitalize the string part
			$capitalized .= firstCharUpperCase($part);
		} // END - foreach

		// Store the converted string in cache array
		$GLOBALS[__FUNCTION__][$str] = $capitalized;
	} // END - if

	// Return cache
	return $GLOBALS[__FUNCTION__][$str];
}

// Generate admin links for mail order
// mailType can be: 'mid' or 'bid'
function generateAdminMailLinks ($mailType, $mailId) {
	// Init variables
	$OUT = '';
	$table = '';

	// Default column for mail status is 'data_type'
	// @TODO Rename column data_type to e.g. mail_status
	$statusColumn = 'data_type';

	// Which mail do we have?
	switch ($mailType) {
		case 'bid': // Bonus mail
			$table = 'bonus';
			break;

		case 'mid': // Member mail
			$table = 'pool';
			break;

		default: // Handle unsupported types
			logDebugMessage(__FUNCTION__, __LINE__, 'Unsupported mail type ' . $mailType . ' for mailId=' . $mailId . ' detected.');
			$OUT = '<div align="center">{%message,ADMIN_UNSUPPORTED_MAIL_TYPE_DETECTED=' . $mailType . '%}</div>';
			break;
	} // END - switch

	// Is the mail type supported?
	if (!empty($table)) {
		// Query for the mail
		$result = SQL_QUERY_ESC("SELECT `id`,`%s` AS `mail_status` FROM `{?_MYSQL_PREFIX?}_%s` WHERE `id`=%s LIMIT 1",
			array(
				$statusColumn,
				$table,
				bigintval($mailId)
			), __FILE__, __LINE__);

		// Do we have one entry there?
		if (SQL_NUMROWS($result) == 1) {
			// Load the entry
			$content = SQL_FETCHARRAY($result);

			// Add output and type
			$content['type']     = $mailType;
			$content['__output'] = '';

			// Filter all data
			$content = runFilterChain('generate_admin_mail_links', $content);

			// Get output back
			$OUT = $content['__output'];
		} // END - if

		// Free result
		SQL_FREERESULT($result);
	} // END - if

	// Return generated HTML code
	return $OUT;
}


/**
 * Determine if a string can represent a number in hexadecimal
 *
 * @param	$hex	A string to check if it is hex-encoded
 * @return	$foo	True if the string is a hex, otherwise false
 * @author	Marques Johansson
 * @link	http://php.net/manual/en/function.http-chunked-decode.php#89786
 */
function isHexadecimal ($hex) {
	// Make it lowercase
	$hex = strtolower(trim(ltrim($hex, '0')));

	// Fix empty strings to zero
	if (empty($hex)) {
		$hex = 0;
	} // END - if

	// Simply compare decode->encode result with original
	return ($hex == dechex(hexdec($hex)));
}

/**
 * Replace "\r" with "[r]" and "\n" with "[n]" and add a final new-line to make
 * them visible to the developer. Use this function to debug e.g. buggy HTTP
 * response handler functions.
 *
 * @param	$str	String to overwork
 * @return	$str	Overworked string
 */
function replaceReturnNewLine ($str) {
	return str_replace("\r", '[r]', str_replace("\n", '[n]
', $str));
}

// Converts a given string by splitting it up with given delimiter similar to
// explode(), but appending the delimiter again
function stringToArray ($delimiter, $string) {
	// Init array
	$strArray = array();

	// "Walk" through all entries
	foreach (explode($delimiter, $string) as $split) {
		//  Append the delimiter and add it to the array
		$strArray[] = $split . $delimiter;
	} // END - foreach

	// Return array
	return $strArray;
}

// Detects the prefix 'mb_' if a multi-byte string is given
function detectMultiBytePrefix ($str) {
	// Default is without multi-byte
	$mbPrefix = '';

	// Detect multi-byte (strictly)
	if (mb_detect_encoding($str, 'auto', true) !== false) {
		// With multi-byte encoded string
		$mbPrefix = 'mb_';
	} // END - if

	// Return the prefix
	return $mbPrefix;
}

// Searches the given array for a sub-string match and returns all found keys in an array
function getArrayKeysFromSubStrArray ($heystack, $needles, $offset = 0) {
	// Init array for all found keys
	$keys = array();

	// Now check all entries
	foreach ($needles as $key => $needle) {
		// Do we have found a partial string?
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'heystack='.$heystack.',key='.$key.',needle='.$needle.',offset='.$offset);
		if (strpos($heystack, $needle, $offset) !== false) {
			// Add the found key
			$keys[] = $key;
		} // END - if
	} // END - foreach

	// Return the array
	return $keys;
}

// Determines database column name from given subject and locked
function determinePointsColumnFromSubjectLocked ($subject, $locked) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'subject=' . $subject . ',locked=' . intval($locked) . ' - ENTERED!');
	// Default is 'normal' points
	$pointsColumn = 'points';

	// Which points, locked or normal?
	if ($locked === true) {
		$pointsColumn = 'locked_points';
	} // END - if

	// Prepare array for filter
	$filterData = array(
		'subject' => $subject,
		'locked'  => $locked,
		'column'  => $pointsColumn
	);

	// Run the filter
	$filterData = runFilterChain('determine_points_column_name', $filterData);

	// Extract column name from array
	$pointsColumn = $filterData['column'];

	// Return it
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'subject=' . $subject . ',locked=' . intval($locked) . ',pointsColumn=' . $pointsColumn . ' - EXIT!');
	return $pointsColumn;
}

// Setter for referral id (no bigintval, or nicknames will fail!)
function setReferralId ($refid) {
	$GLOBALS['refid'] = $refid;
}

// Checks if 'refid' is valid
function isReferralIdValid () {
	return ((isset($GLOBALS['refid'])) && (getReferralId() !== NULL) && (getReferralId() > 0));
}

// Getter for referral id
function getReferralId () {
	return $GLOBALS['refid'];
}

// Converts a boolean variable into 'Y' for true and 'N' for false
function convertBooleanToYesNo ($boolean) {
	// Default is 'N'
	$converted = 'N';
	if ($boolean === true) {
		// Set 'Y'
		$converted = 'Y';
	} // END - if

	// Return it
	return $converted;
}

// Translates task type to a human-readable version
function translateTaskType ($taskType) {
	// Construct message id
	$messageId = 'ADMIN_TASK_TYPE_' . strtoupper($taskType) . '';

	// Is the message id there?
	if (isMessageIdValid($messageId)) {
		// Then construct message
		$message = '{--' . $messageId . '--}';
	} else {
		// Else it is an unknown task type
		$message = '{%message,ADMIN_TASK_TYPE_UNKNOWN=' . $taskType . '%}';
	} // END - if

	// Return message
	return $message;
}

// Translates points subject to human-readable
function translatePointsSubject ($subject) {
	// Construct message id
	$messageId = 'POINTS_SUBJECT_' . strtoupper($subject) . '';

	// Is the message id there?
	if (isMessageIdValid($messageId)) {
		// Then construct message
		$message = '{--' . $messageId . '--}';
	} else {
		// Else it is an unknown task type
		$message = '{%message,POINTS_SUBJECT_UNKNOWN=' . $subject . '%}';
	} // END - if

	// Return message
	return $message;
}

// "Translates" 'true' to true and 'false' to false
function convertStringToBoolean ($str) {
	// Trim it lower-case for validation
	$str = trim(strtolower($str));

	// Is it valid?
	if (!in_array($str, array('true', 'false'))) {
		// Not valid!
		debug_report_bug(__FUNCTION__, __LINE__, 'str=' . $str . ' is not true/false');
	} // END - if

	// Return it
	return (($str == 'true') ? true : false);
}

/**
 * "Makes" a variable in given string parseable, this function will throw an
 * error if the first character is not a dollar sign.
 *
 * @param	$varString	String which contains a variable
 * @return	$return		String with added single quotes for better parsing
 */
function makeParseableVariable ($varString) {
	// The first character must be a dollar sign
	if (substr($varString, 0, 1) != '$') {
		// Please report this
		debug_report_bug(__FUNCTION__, __LINE__, 'varString=' . $varString . ' - No dollar sign detected, will not parse it.');
	} // END - if

	// Do we have cache?
	if (!isset($GLOBALS[__FUNCTION__][$varString])) {
		// Snap them in, if [,] are there
		$GLOBALS[__FUNCTION__][$varString] = str_replace('[', "['", str_replace(']', "']", $varString));
	} // END - if

	// Return cache
	return $GLOBALS[__FUNCTION__][$varString];
}

//-----------------------------------------------------------------------------
// Automatically re-created functions, all taken from user comments on www.php.net
//-----------------------------------------------------------------------------
if (!function_exists('html_entity_decode')) {
	// Taken from documentation on www.php.net
	function html_entity_decode ($string) {
		$trans_tbl = get_html_translation_table(HTML_ENTITIES);
		$trans_tbl = array_flip($trans_tbl);
		return strtr($string, $trans_tbl);
	}
} // END - if

// [EOF]
?>
